

class FlatRecordBuilder(object):
    """A builder of FlatRecords.
    
    A 'FlatRecord' is a Record whose keys are tuples. The keys of nested keys will be 'flatted'.
    For example:

    {
        "field_1": 5,
        "field_2": {
            "field_x": 7,
            "field_y": 9
        }
    }

    The FlatRecord will be:
    {
        ("field_1",): 5,
        ("field_2","field_x"): 7,
        ("field_2","field_y"): 9
    }
    """

    @staticmethod
    def get_flat_record_from_normal_record(record: dict) -> dict:
        """Flattens a record.

        :param record: The input record.
        :type record: dict
        :return: A flat record.
        :rtype: dict
        """
        res = {}
        for key, value in record.items():
            flat_key = (key,)
            if isinstance(value, dict):
                subrecord = FlatRecordBuilder.get_flat_record_from_normal_record(value)
                res.update({flat_key+flat_subkey: subvalue for flat_subkey, subvalue in subrecord.items()})
            else:
                res.update({flat_key: value})

        return res 
    
    @staticmethod
    def get_normal_record_from_flat_record(flat_record: dict) -> dict:
        """
        Return a "normal record" from a FlatRecord.
        :param flat_record: A flat record.
        :type flat_record: dict
        :return: A normal record.
        :rtype: dict
        """

        res_record = {}
        for composed_key, value in flat_record.items():
            if len(composed_key) == 1:
                res_record[composed_key[0]] = value

            elif len(composed_key) == 2:
                super_key, nested_key = composed_key
                if super_key not in res_record:
                    res_record[super_key] = {}
                
                if isinstance(res_record[super_key], dict):
                    res_record[super_key][nested_key] = value
                else:
                    raise RuntimeError(f"Nested key in a non-dict field: {composed_key}")
            else:
                raise NotImplementedError(f"Three-depth record level is not supported -> {composed_key}")
        
        return res_record


import csv
import json
from typing import BinaryIO, List, Iterable

from RecordMapper.common import Writer


class CSVWriter(Writer):
    """The object that writes records to a csv file.
    """

    def __init__(self, file_path: str, fieldnames: List[str]):
        """The constructor of the CSVWriter.

        :param file_path: The path to write.
        :type file_path: object.
        :param fieldnames: The list of names of the columns.
            Their order will be preserved in the result file.
        :type fieldnames: List[str]
        """

        super().__init__(file_path)
        self.fieldnames = fieldnames
        self.write_count = 0

    def write_records_to_output(self, records: Iterable[dict], output: BinaryIO, output_opts: dict):
        """Writes the records to an output file.

        :param records: An iterable of records.
        :type records: List[dict]
        :param output: The object to write.
        :type output: BinaryIO
        """

        fields_to_flat = set(output_opts.get("flat_nested_schema_on_csv", dict()).keys())

        available_fieldnames = [
            fieldname
            for fieldname in self.fieldnames
            if fieldname not in fields_to_flat
        ]

        csv_writer = csv.DictWriter(output, fieldnames=available_fieldnames)
        csv_writer.writeheader()

        for record in records:
            csv_writer.writerow(self.format_record(record, fields_to_flat))
            self.write_count += 1

    def format_record(self, record: dict, fields_to_flat: set) -> dict:
        """Format a record to be written. For example, dicts will be serialized to
        JSON strings.

        :param record: A input record.
        :type record: dict
        :param fields_to_flat: A set of fields that will be flatted.
        :type flatten: set
        :return: A formatted record to be written.
        :rtype: dict
        """

        record_to_write = {**record}

        for key, value in record.items():
            if isinstance(value, dict):
                if key in fields_to_flat:
                    record_to_write = {**record_to_write, **value}
                    del record_to_write[key]
                else:
                    record_to_write[key] = json.dumps(value)

        return record_to_write


import re
import importlib
from collections import namedtuple
from inspect import getmembers, isfunction
from typing import Callable, List


class TransformApplier(object):
    """This applier executes the defined transformations for each field.
    """

    def __init__(self, custom_variables: dict):
        """The constructor of the applier. It accepts a 'custom_variables' argument that its value
        will be passed to functions.
        :param custom_variables: A dict of custom variables.
        :type custom_variables: dict
        """

        self.custom_variables = custom_variables

    def apply(self, flat_record: dict, flat_schema: dict) -> (dict, dict):
        """Executes the function of this applier. It executes the defined functions in "transform"
        value in several phases. Each phase receives the resulting record of the previous one (except the first
        phase, that receives the initial record)-

        :param flat_record: The input record.
        :type flat_record: dict
        :param flat_schema: The input flat schema.
        :type flat_schema: dict
        return: The transformed record and the original flat schema.
        :rtype: (dict, dict)
        """

        # The maximum of transform phases to perform
        max_phases = max([len(field_data.transforms) for _, field_data in flat_schema.items()])
        # Initial copy of the record
        new_record = {**flat_record} 

        # Iterate over phases and get the transform_functions
        for phase_index in range(max_phases):
            new_values_in_this_phase = {}
            transforms_by_key = [
               (field_key, field_data.transforms[phase_index])
               for field_key, field_data in flat_schema.items()
               if phase_index < len(field_data.transforms) and field_data.transforms[phase_index] is not None
            ] 

            # Iterate over the transform functions
            for key, transform_function in transforms_by_key:
                res = transform_function(new_record.get(key), new_record, flat_schema, self.custom_variables)

                # If res is a single value, it modifies only the current value
                if type(res) is not tuple:
                    new_values_in_this_phase[key] = res
                # If res is a tuple of two values and the second one is a dict, it adds the first one as usually
                # and the second one as update. 
                elif type(res) is tuple and len(res) == 2 and type(res[1]) is dict:
                    new_values_in_this_phase[key] = res[0]
                    new_values_in_this_phase.update(res[1])
                # Otherwise, it is not a valid result
                else:
                    raise RuntimeError(f"Invalid result in a transform function of the key: '{key}'")
            
            new_record = {**new_record, **new_values_in_this_phase}

        return new_record, flat_schema 


from typing import Callable, List

def chain_functions(*functions_list: List[Callable]) -> Callable:
    """A helper function that creates another one that executes all the functions
    defined in functions_list in a waterfall way. 
    :param functions_list: A list of functions.
    """

    def wrapper_function(*input_args: List[object]) -> tuple:

        res = input_args

        for single_function in functions_list:
            args_as_list = object_as_tuple(res)
            res = single_function(*args_as_list) 

        return res
    
    return wrapper_function


def object_as_tuple(obj: object) -> tuple:
    """Transforms an object into a tuple (except if this object is already a tuple).

    :param obj: An object 
    :type obj: object
    :return: 
    :rtype: [type]
    """

    if isinstance(obj, tuple):
        return obj
    else:
        return (obj,)


from typing import List, Union

from collections import namedtuple

class RenameApplier(object):
    """An applier that applies a renaming process using the defined aliases on the schema
    using the avro convention"""

    def __init__(self, custom_variables: dict):
        """The constructor of the applier.
        :param custom_variables: A dict of custom variables.
        :type custom_variables: dict
        """

        self.custom_variables = custom_variables

    def apply(self, flat_record: dict, flat_schema: dict) -> (dict, dict):
        """Executes the function of this applier. It renames the fields of the input record
        using the "aliases" value of each field and following the Avro convention.
        IMPORTANT: The "rename process" creates new fields, but it doesn't remove the old ones.
        So, if a record is {'field_1': 5} and the schema has the definition
        {"name": "field_2", "aliases": ["field_1"], "type": "int"}, the resulting record will be
        {"field_1": 5, "field_2": 5}.

        :param flat_record: The input flat record.
        :type flat_record: dict
        :param flat_schema: The input flat schema.
        :type flat_schema: dict
        :return: The record with the "renamed" fields.
        :rtype: dict
        """

        new_record = {**flat_record}

        # Rename each file of the original record
        for key, field_data in flat_schema.items():

            for alias in field_data.aliases:
                # As tuple key
                if (alias,) in new_record:
                    new_record[key] = new_record[(alias,)]

        return (new_record, flat_schema)



from typing import BinaryIO, Iterator

import fastavro

from RecordMapper.common import Reader

class AvroReader(Reader):
    """A Record reader from Avro format.
    """

    def __init__(self, path_to_read: str):
        """Constructor of the AvroRecord.

        :param path_to_read: Path of the file to read.
        :type path_to_read: string
        """

        super().__init__(path_to_read)

        self.read_options ="rb"

    def read_records_from_input(self, input_stream: BinaryIO) -> Iterator[dict]:
        """Read records from an input stream in avro format.

        :param input_stream: The input stream.
        :type input_stream: BinaryIO
        :yield: A record.
        :rtype: Iterator[dict]
        """

        for record in fastavro.reader(input_stream):
            yield record




from typing import Dict
from RecordMapper.builders.FlatSchemaBuilder import FieldData

class NestedSchemaSelectorApplier(object):
    """An applier that modifies a FlatSchema selecting multiple nested schemas using
       the nestedSchemaSelector function.
    """

    def __init__(self, flat_schemas: Dict[str, dict], custom_variables: dict):
        """The constructor function of this class.

        :param flat_schemas: A list of FlatSchemas.
        :type flat_schemas: Dict[str, dict]
        :param custom_variables: A dict of custom variables.
        :type custom_variables: dict
        """

        self.flat_schemas = flat_schemas
        self.custom_variables = custom_variables

    def apply(self, record: dict, base_flat_schema: dict) -> (dict, dict):
        """Execute the function of this applier. If a field has several nested schemas as type, this applier
        creates new base_flat_schema selecting one of the nested schemas (adding their fields to the resulting flat schema)
        using the function defined in the value "nestedSchemaSelector".

        :param record: The input record. 
        :type record: dict
        :param base_flat_schema: The input base flat schema. 
        :type base_flat_schema: dict. 
        :return: The record (unmodified) and the flat schema (modified with the selected nested schemas).
        :rtype: (dict, dict)
        """

        # The value that will be returned with the record
        complete_flat_schema = {**base_flat_schema}

        for field_key, field_data in base_flat_schema.items():

            if field_data.selector is not None:
                complete_flat_schema = self.select_nested_schema_and_add_their_fields(record, complete_flat_schema,
                                                                                      field_key, field_data)
                # Remove the "super key" to avoid problems
                del complete_flat_schema[field_key]

        return (record, complete_flat_schema)

    def select_nested_schema_and_add_their_fields(self, record: dict, flat_schema: dict, field_key: str,
                                                  field_data: FieldData) -> dict:
        """Select a nested schema for a field (using the defined function in "selector" value of the FieldData) and add their fields
        to the resulting flat schema.

        :param record: The input record. It will be used as argument
        :type record: dict
        :param flat_schema: The input flat schema.
        :type flat_schema: dict
        :param field_key: The key to "analyze".
        :type field_key: str
        :param field_data: The FieldData object of the previous key.
        :type field_data: tuple
        :raises RuntimeError: This function will raise if the selected nested schema name is not valid. 
        :return: The modified flat Schema.
        :rtype: dict
        """

        nested_schema_name = field_data.selector(field_key, record, self.custom_variables) \
            if field_data.selector != None and field_data.selector(field_key, record, self.custom_variables) != None else None

        # Not schema selected (this field will be null)
        if nested_schema_name == None:
            return flat_schema

        # Schema selected. Add their fields to the complete schema
        elif nested_schema_name in self.flat_schemas:

            nested_schema = self.flat_schemas[nested_schema_name]

            fields_to_add = dict(
                [
                    (field_key + nested_key, nested_field_data)
                    for nested_key, nested_field_data in nested_schema.items()
                ]
            )

            return {**flat_schema, **fields_to_add}
        else:
            raise RuntimeError(f"Invalid nested schema name: {nested_schema_name}")



from typing import BinaryIO, Iterator


class Reader(object):
    """A generic Reader object.
    """

    def __init__(self, path_to_read: str):
        """The constructor function.

        :param path_to_read: The path to read. 
        :type path_to_read: str
        """

        self.file_path = path_to_read
        self.read_options = "r"
        self.input_stream = None

    def read_records(self) -> Iterator[dict]:
        """Open the file and read the records.  

        :return: Returns a generator of records.
        :rtype: Iterator[dict] 
        """

        self.input_stream = open(self.file_path, self.read_options)

        return self.read_records_from_input(self.input_stream)

    def read_records_from_input(self, input_stream: BinaryIO) -> Iterator[dict]:
        """This function reads from the input object and returns a generator of records.
        This has to be implemented by a child class.

        :param input_stream: An input stream.
        :type input_stream: BinaryIO.
        :raises NotImplementedError: This function has to be implemented by a child class.
        :yield: A read record.
        :rtype: Iterator[dict]
        """

        raise NotImplementedError("This method has to be implemented to read formatted records!")

    def close(self):
        """This method close the input stream.
        """

        self.input_stream.close()





class CleanApplier(object):
    """An applier that removes the unused files of a record (using a schema as a reference).
    """

    def __init__(self, custom_variables: dict):
        """The constructor of the applier.
        :param custom_variables: A dict of custom variables.
        :type custom_variables: dict
        """

        self.custom_variables = custom_variables

    def apply(self, flat_record: dict, flat_schema: dict) -> (dict, dict):
        """The apply of the CleanApplier. Removed unused files in a FlatRecord using
        a FlatSchema as reference (remove the fields that are not in the schema).

        :param flat_record: The input FlatRecord.
        :type flat_record: dict
        :param flat_schema: A FlatSchema. 
        :type flat_schema: dict
        :return: A clean FlatRecord and the original FlatSchema.
        :rtype: (dict,dict)
        """
        
        new_record = {}

        for key, field_data in flat_schema.items():

            if key in flat_record:
                new_record[key] = flat_record[key]
        
        return new_record, flat_schema
 


from typing import Union, List, Callable
from collections import namedtuple

from RecordMapper.builders import FunctionBuilder

FieldData = namedtuple("FieldData", ["types", "aliases", "transforms", "selector"])

class FlatSchemaBuilder(object):
    """A builder class that creates a FlatSchema.
    """

    @staticmethod
    def get_flat_schema(schema: dict) -> dict:
        """Creates a FlatSchema from a normal avro Schema.

        A FlatSchema is a dict where the keys are tuples and the values "FieldData" objects.

        :param schema: Normal Avro schema. 
        :type schema: dict
        :return: A FlatSchema.
        :rtype: dict
        """

        schema_fields = schema["fields"]

        flat_schema = {}

        for field in schema_fields:

            field_name = field["name"]
            field_types = field["type"] if not isinstance(field["type"], str) else [field["type"]]
            field_aliases = field["aliases"] if "aliases" in field else []
            field_transforms = FlatSchemaBuilder.get_transform_functions_from_field(field.get("transform", None)) 
            field_selector = FlatSchemaBuilder.get_selector_function(field.get("nestedSchemaSelector", None))

            field_data = FieldData(field_types, field_aliases, field_transforms, field_selector) 

            # Key as tuple
            flat_schema[(field_name,)] = field_data
        
        return flat_schema

    @staticmethod
    def get_transform_functions_from_field(transform_field: Union[str, List[str]]) -> List[Callable]:
        """Generate a list of transform functions from a list of string values (or a single string value).

        :param transform_field: A list of string values (or a single string value) that reference one or several transform functions.
        :type transform_field: Union[str, List[str]]
        :return: List of transform functions.
        :rtype: List[Callable]
        """

        # First, get a string list 
        function_str_list = None
        if transform_field is None:
            function_str_list = []
        elif isinstance(transform_field, str):
            function_str_list = [transform_field]
        else:
            function_str_list = transform_field

        return [FunctionBuilder.parse_function_str(function_str) for function_str in function_str_list]

    @staticmethod
    def get_selector_function(selector_field: str) -> Callable:
        """Generates a Selector function from a string.

        :param selector_field: A string that represents a Selector function.
        :type selector_field: str
        :return: A Selector function.
        :rtype: Callable
        """

        if selector_field is None:
            return None
        else:
            return FunctionBuilder.parse_function_str(selector_field)



from typing import BinaryIO, Iterable


class Writer(object):
    """The generic writer class.
    """

    def __init__(self, file_path: str):
        """The constructor function.

        :param file_path: The path of the input file. 
        :type file_path: str
        """

        self.file_path = file_path
        self.write_options = "w"

        self.output_stream = None

    def write_records(self, records: Iterable, output_opts: dict):
        """This function writes the records to the output file.

        :param records: An iterable of records.
        :type records: Iterable
        :return: A generator of records.
        :rtype: Iterator[dict]
        """

        self.output_stream = open(self.file_path, self.write_options)

        return self.write_records_to_output(records, self.output_stream, output_opts)

    def write_records_to_output(self, records: Iterable, output: BinaryIO, output_opts: dict):
        """Write the records to an output file. This function has to be implemented by a
        child class.

        :param records: An iterable of records.
        :type records: Iterable
        :param output: An output stream.
        :type output: BinaryIO
        :raises NotImplementedError: This function has to be implemented by a child class.
        """

        raise NotImplementedError("This method has to be implemented to write formatted records!")

    def close(self):
        """This function close the output stream.
        """

        self.output_stream.close()



import csv
from typing import BinaryIO, Iterator

from RecordMapper.common import Reader


csv.field_size_limit(100_000_000)


class CSVReader(Reader):
    """A object that reads records form an input csv file.
    """

    def __init__(self, file_path: str):
        """The constructor of the CSVReader.

        :param input: The path of the file.
        :type input: str
        """

        super().__init__(file_path)
        self.filepath = file_path

    def read_records_from_input(self, input_stream: BinaryIO) -> Iterator[dict]:
        """The function that reads the records from the input stream.

        :param input_stream: The input stream of the records.
        :type input_stream: BinaryIO
        :yield: dict
        :rtype: 
        """

        self.reader = csv.DictReader(input_stream)

        for record in self.reader:
            formatted_record = dict([(key, value if value != '' else None) for key, value in record.items()])
            yield formatted_record



import itertools
from typing import List, Iterable, Iterator

from RecordMapper.appliers import NestedSchemaSelectorApplier, RenameApplier, TransformApplier, CleanApplier
from RecordMapper.avro.AvroReader import AvroReader
from RecordMapper.avro.AvroWriter import AvroWriter
from RecordMapper.builders import FlatSchemaBuilder, FlatRecordBuilder
from RecordMapper.csv.CSVReader import CSVReader
from RecordMapper.csv.CSVWriter import CSVWriter
from RecordMapper.xml.XMLReader import XMLReader
from RecordMapper.utils import chain_functions


class RecordMapper(object):
    """The main class of this package. It has methods to transform data using an Avro Schema (and custom functions).
    """

    def __init__(self, base_schema: dict, nested_schemas: List[dict] = [], custom_variables: dict = {}):
        """__init__ function of RecordMapper.

        :param base_schema: The base schema in avro format to transform the records.
        :type base_schema: dict
        :param nested_schemas: The schemas of the nested record types. Defaults to [].
        :type nested_schemas: List[dict], optional
        :param custom_variables: A dict of custom variables that will be accesible for the appliers.
        """
        self.original_base_schema = base_schema
        self.original_nested_schemas = nested_schemas
        self.custom_variables = custom_variables

        # Stats of the record mapper
        self.stats = {}

        self.flat_schemas = dict(
            [
                (schema["name"], FlatSchemaBuilder.get_flat_schema(schema))
                for schema in ([self.original_base_schema] + self.original_nested_schemas)
            ]
        )

        self.selector_applier = NestedSchemaSelectorApplier(self.flat_schemas, self.custom_variables)
        self.rename_applier = RenameApplier(self.custom_variables)
        self.transform_applier = TransformApplier(self.custom_variables)
        self.clean_applier = CleanApplier(self.custom_variables)

    def execute(self, input_format: str, input_file_path: str, paths_to_write: dict,
                input_opts: dict = {}, base_schema_to_write: dict = None, nested_schemas_to_write: List[dict] = None,
                output_opts: dict = {}):
        """Process that reads, transforms a writes the data from a format to another one.

        :param input_format: The input format (csv, avro, ...)
        :type input_format: str
        :param input_file_path: The path of the input file.
        :type input_file_path: str
        :param paths_to_write: A dict with the output formats (avro, csv...) and their configuration.
        :type paths_to_write: dict
        :param input_opts: Some special options of the input process, defaults to {}
        :type input_opts: dict, optional
        :param base_schema_to_write: A base schema used to write, defaults to None
        :type base_schema_to_write: dict, optional
        :param nested_schemas_to_write: A list of nested record schemas used to write, defaults to None
        :type nested_schemas_to_write: List[dict], optional
        """
        
        # Reset stats dict
        self.stats = {}

        read_records = self.read_records(input_format, input_file_path, input_opts)
        transformed_records = self.transform_records(read_records)
        self.write_records(transformed_records, paths_to_write, base_schema_to_write, nested_schemas_to_write,
                           output_opts)

    def transform_records(self, record_list: Iterable[dict]) -> Iterator[dict]:
        """A generator of transformed records.

        :param record_list: An iterable of records.
        :type record_list: Iterable[dict]
        :yield: A transformed record.
        :rtype: Iterator[dict]
        """

        for record in record_list:
            yield self.transform_record(record)

    def transform_record(self, record: dict) -> dict:
        """Apply transforms on a single record. Each transform step is defined in an Applier object.

        :param record: An input record.
        :type record: dict
        :return: A transformed record.
        :rtype: dict
        """

        base_schema_name = self.original_base_schema["name"]
        base_flat_schema = self.flat_schemas[base_schema_name]

        flat_record = FlatRecordBuilder.get_flat_record_from_normal_record(record)

        all_functions = chain_functions(
            self.selector_applier.apply,
            self.rename_applier.apply,
            self.transform_applier.apply,
            self.clean_applier.apply
        )

        transformed_record, transformed_flat_schema = all_functions(flat_record, base_flat_schema)

        normal_record = FlatRecordBuilder.get_normal_record_from_flat_record(transformed_record)

        return normal_record

    def read_records(self, input_format: str, path_to_read: str, opts: dict = {}) -> Iterator[dict]:
        """Read records of an input file.

        :param input_format: The format of the file.
        :type input_format: str
        :param path_to_read: The path of the file.
        :type path_to_read: str
        :param opts: Some special options in read process, defaults to {}
        :type opts: dict, optional
        :raises RuntimeError: [description]
        :return: [description]
        :rtype: Iteraror[dict]
        :yield: [description]
        :rtype: Iterator[Iteraror[dict]]
        """

        self.stats["read_count"] = 0

        if input_format == "avro":
            reader_object = AvroReader(path_to_read)
        elif input_format == "csv":
            reader_object = CSVReader(path_to_read)
        elif input_format == "xml":
            reader_object = XMLReader(path_to_read)
        else:
            raise RuntimeError(f"Invalid input format: {input_format}")

        for record in reader_object.read_records():
            self.stats["read_count"] += 1
            yield record

        reader_object.close()

    def write_records(self, records_list: Iterable, paths_to_write: dict, base_schema_to_write: dict = None,
                      nested_schemas_to_write: List[dict] = None, output_opts: dict = {}):
        """Writes records to one or several formats. It is necessary to write to an avro file at least.

        :param records_list: An iterable of records.
        :type records_list: Iterable
        :param paths_to_write: A dict with the formats and the paths that will be written. For example, {"csv": "example.csv", "avro": "example.avro"}.
        :type paths_to_write: dict
        :param base_schema_to_write: A base schema used in the write process different from the one
            used in the transform process, defaults to None
        :type base_schema_to_write: dict, optional
        :param nested_schemas_to_write: A list of nested schemas used in the write process different
            from the one used in the transform process, defaults to None
        :type nested_schemas_to_write: List[dict], optional
        :raises RuntimeError: Raises and error if there is not a specified path for the avro format.
        """

        self.stats["write_count"] = {}

        base_schema_to_write = self.original_base_schema if base_schema_to_write is None else base_schema_to_write
        nested_schemas_to_write = self.original_nested_schemas if nested_schemas_to_write is None else nested_schemas_to_write

        if "avro" not in paths_to_write:
            raise RuntimeError("It is necessary a path to write an Avro File")

        # Avro writing
        writer_avro = AvroWriter(
            paths_to_write["avro"],
            base_schema_to_write,
            nested_schemas_to_write,
            output_opts
        )
        writer_avro.write_records(records_list, output_opts)
        self.stats["write_count"]["avro"] = writer_avro.write_count
        writer_avro.close()

        if "csv" in paths_to_write:

            fieldnames = [field["name"] for field in base_schema_to_write["fields"]]

            ## If we want to consider to flatten the nested schemas:
            if output_opts.get("flat_nested_schema_on_csv", None) != None:
                keys = output_opts["flat_nested_schema_on_csv"].values()
                fieldnames_nested = [nested_schema['fields'] for nested_schema in nested_schemas_to_write if
                                     nested_schema['name'] in keys]

                fields_concatenated = list(itertools.chain.from_iterable(fieldnames_nested))

                fieldnames_concatenated = [entry['name'] for entry in fields_concatenated]

                fieldnames += fieldnames_concatenated

            records = AvroReader(paths_to_write["avro"]).read_records()
            writer_csv = CSVWriter(paths_to_write["csv"], fieldnames)
            writer_csv.write_records(records, output_opts)
            self.stats["write_count"]["csv"] = writer_csv.write_count



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['RecordMapper',
 'RecordMapper.appliers',
 'RecordMapper.avro',
 'RecordMapper.builders',
 'RecordMapper.common',
 'RecordMapper.csv',
 'RecordMapper.xml']

package_data = \
{'': ['*']}

install_requires = \
['coverage>=4.5.4,<4.6.0',
 'dateparser>=1.0.0,<1.1.0',
 'defusedxml>=0.6.0,<0.7.0',
 'fastavro>=0.23.4,<0.24.0',
 'nose>=1.3.7,<1.4.0']

setup_kwargs = {
    'name': 'recordmapper',
    'version': '0.6.3',
    'description': 'Transform records using an Avro Schema and custom map functions',
    'long_description': '# RecordMapper\nRead, transform and write records using an Avro Schema and custom map functions.\n\n## Update PyPI version\n\n    poetry publish --build',
    'author': 'UDARealState Data engineering Team',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/urbandataanalytics/RecordMapper',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
This module defines a set of built-in functions.
"""
import math
from datetime import datetime
from typing import Union, List, Iterable

import dateparser


def copyFrom(path_to_copy_from: str):
    """This built-in function returns the value of 'path_to_copy_from' key.
    """

    def transform_function(current_value: object, record: dict, complete_transform_schema: dict,
                           custom_variables: dict):
        composed_key = (path_to_copy_from,)
        return record.get(composed_key, None)

    return transform_function


def toNull():
    """This returns always null.
    """

    def transform_function(current_value: object, record: dict, complete_transform_schema: dict,
                           custom_variables: dict):
        return None

    return transform_function


def toFloat():
    """This built-in function casts the current value to Float and returns the result.
    """

    def transform_function(current_value: object, record: dict, complete_transform_schema: dict,
                           custom_variables: dict):
        value_to_return = None

        if current_value is not None:
            try:
                value_to_return = float(current_value)
            except:
                value_to_return = None

        return value_to_return

    return transform_function


def toInt():
    """This built-in function casts the current value to Int and returns the result.
    """

    def transform_function(current_value: object, record: dict, complete_transform_schema: dict,
                           custom_variables: dict):
        value_to_return = None

        if current_value is not None:
            try:
                clean_current_value = current_value.replace(",", ".")
                float_current_value = float(clean_current_value)
                value_to_return = int(math.floor(float_current_value))
            except:
                value_to_return = None

        return value_to_return

    return transform_function


def toRoundedInt(use_floor: bool):
    """This built-in function casts the current value to Int and returns the result.
    :param floor: if use the floor method to truncate the value or to use the ceiling if floor is not given
    :type floor: str
   """

    def transform_function(current_value: object, record: dict, complete_transform_schema: dict,
                           custom_variables: dict):
        value_to_return = None

        if current_value is not None:
            try:
                float_current_value = float(current_value)
                value_to_return = int(math.floor(float_current_value)) if use_floor else int(math.ceil(float_current_value))
            except:
                value_to_return = None

        return value_to_return

    return transform_function


def toString():
    """This built-in function casts the current value to String and returns the result.
    """

    def transform_function(current_value: object, record: dict, complete_transform_schema: dict, custom_variables: dict,
                           is_nested_record: bool = False):
        if current_value is None:
            return None
        else:
            return str(current_value)

    return transform_function


def toBool(value_for_true: str):
    """This built-in function casts the current value to Bool and returns the result.
    """

    def transform_function(current_value: object, record: dict, complete_transform_schema: dict, custom_variables: dict,
                           is_nested_record: bool = False):
        if current_value is None:
            return None
        else:
            return value_for_true == current_value

    return transform_function


def toDate(*formats: Iterable[str]):
    """This built-in function casts the current value to Date and returns the result.

    :param formats: the passed format which will guide the parser to build the datetime correctly
    :type formats: Union[str, List[str]]
    """

    def transform_function(current_value: object, record: dict, complete_transform_schema: dict, custom_variables: dict,
                           is_nested_record: bool = False):
        if current_value is None:
            return None
        else:
            for format in formats:
                try:
                    date_time_obj = datetime.strptime(current_value, format)
                except ValueError:
                    pass
                else:
                    return str(date_time_obj)

            return None

    return transform_function


def timestampToDate(format: str):
    """This built-in function casts the current value to Date and returns the result.

    :param format: the passed format which will guide the parser to build the datetime correctly
    :type format: str
    """

    def transform_function(current_value: object, record: dict, complete_transform_schema: dict, custom_variables: dict,
                           is_nested_record: bool = False):
        if current_value is None:
            return None
        else:
            dt = datetime.fromtimestamp(float(current_value) / 1000)

            formatted_time = dt.strftime(f'{format}')

            return str(formatted_time)

    return transform_function


def get_from_custom_variable(variable_name: str):
    """This built-in function gets from custom_variables the value 'variable_name'.

    :param variable_name: The defined variable_name.
    :type variable_name: str
    """

    def transform_function(current_value: object, record: dict, complete_transform_schema: dict,
                           custom_variables: dict):
        return custom_variables[variable_name]

    return transform_function


def transform_date_between_formats(input_format: str, output_format: str):
    """This built-in function transform given date as str from an input_format to an output_format.

    If given input date does not match input format, it will try to parse input date with dateparser.
    You can force dateparser parsing giving any incorrect input_format like "ignore", anyway  if
    a correct format does not match the given date, it will use dateparser.
    """

    def transform_function(current_value: str, record: dict, complete_transform_schema: dict, custom_variables: dict,
                           is_nested_record: bool = False):
        if current_value:
            try:
                datetime_before_transform = datetime.strptime(current_value, input_format)
            except ValueError:
                datetime_before_transform = dateparser.parse(current_value)

                if datetime_before_transform > datetime.now():
                    datetime_before_transform = datetime_before_transform.replace(year=datetime_before_transform.year-1)

            return datetime.strftime(datetime_before_transform, output_format)

    return transform_function



from typing import BinaryIO, List, Iterable

import fastavro

from RecordMapper.common import Writer


class AvroMatchingException(Exception):
    """
        Exception when a row is writed to an avro file and it doesn't match with the schema.
    """
    pass


class AvroWriter(Writer):
    """This object create a writer that writes avro data into a file-like object.
    """

    def __init__(self, obj_to_write: object, base_schema: dict, nested_schemas: List[dict] = [], output_opts: dict = {}):
        """AvroWriter constructor

        :param output_stream: The file-like object where data will be writed.
        :type output_stream: file
        :param avro_schema: A valid avro schema as a dict.
        :type avro_schema: dict
        """

        super().__init__(obj_to_write)
        self.write_options = "wb"
        self.write_count = 0

        self.parsed_nested_schemas = [fastavro.parse_schema(schema) for schema in nested_schemas]

        if output_opts.get('merge_schemas', False):
            self.parsed_base_schema = self.merge_schemas(base_schema, nested_schemas)

        else:
            self.parsed_base_schema = fastavro.parse_schema(base_schema)



    def write_records_to_output(self, record_list: Iterable, output: BinaryIO, output_opts: dict):

        # As fastavro is not able to use nested schemas, we must combine the base schema and the nested schema
        self.writer = fastavro.write.Writer(output, self.parsed_base_schema)

        for record in record_list:
            try:
                self.writer.write(record)
                self.write_count += 1
            except ValueError as ex:
                raise AvroMatchingException(f"Exception: {ex} for row -> {record}")

    def close(self):
        """Sends the buffer reamining data and closes the output stream
        """

        self.writer.flush()
        super().close()

    def merge_schemas(self, base_schema: dict, nested_schemas: List[dict]):

        nested_schemas_dict = {nested_schema['name']: nested_schema for nested_schema in nested_schemas}

        res_schema = {**base_schema}

        for field in res_schema['fields']:
            type_list = field['type'] if isinstance(field['type'], list) else [field['type']]

            field['type'] = [nested_schemas_dict[value] if value in nested_schemas_dict else value
                             for value in type_list]

        return res_schema


import re
from typing import Callable, List, Union
from inspect import getmembers, isfunction
import importlib

from RecordMapper.builders import BuiltinFunctions


class InvalidFunctionError(Exception):
    """An exception that represents an invalid string reference for a function"""
    pass


class FunctionBuilder(object):
    """A Builder class of functions"""

    @staticmethod
    def parse_function_str(function_str: str) -> Union[Callable[[object, dict], object], None]:
        """
        This function parses a string and generates a function to be used as a transform function.
        If function_str is None, it returns None.
        :param function_str: A string that represents a function.
        "type function_str: str
        :return: A transform function.
        :rtype: Union[Callable[[object, dict], object], None]
        """

        if function_str is None:
            return None

        parsed_function = re.match("^([\.\w]+)(?:\(([\w|,%\'-: ]*)\))?$", function_str)
        if not parsed_function:
            raise RuntimeError(f"Invalid name for a transform function: '{function_str}'")

        function_name, args = parsed_function.groups()
        args_list = str(args).split(",") if (args is not None and args != '') else []

        # Check if it is a built-in function
        builtin_function = FunctionBuilder.get_builtin_function(function_name, args_list)

        if builtin_function is not None:
            return builtin_function

        # Get it as custom function
        return FunctionBuilder.get_custom_function(function_name, args_list)

    @staticmethod
    def get_builtin_function(function_name: str, args_list: List[str]) -> Callable[[object, dict], object]:
        """Get a built-in function from a string.

        :param function_name: A string that represents a transform function.
        :type function_name: str
        :param args_list: Argument list for the function (that will return another function when it is executed).
        :type args_list: List[str]
        :return: A transform function.
        :rtype: Callable[[object, dict], object]

        """
        # Check if it is a built-in function
        possible_function = [obj for name, obj in getmembers(BuiltinFunctions) if
                             name == function_name and isfunction(obj)]

        if len(possible_function) == 1:
            return possible_function[0](*args_list)
        else:
            return None

    @staticmethod
    def get_custom_function(function_name: str, args_list: List[str]):
        """Get a custom function from a string. This string is a complete import path to the function.
        For example, example1.example2.example_function references the "example_function" in the module
        example2 inside the package example1.

        :param function_name: The name of the function (the complete import path).
        :type function_name: str
        :param args_list: Argument list for the function.
        :type args_list: List[str]
        :raises InvalidFunctionError: It raises an exception when the import path of the function is invalid.
        :return: A custom Transform function. 
        :rtype: Callable[[object, dict], object]
        """

        parts = function_name.split(".")
        module_path = ".".join(parts[:-1])
        function_name = parts[-1]
        try:
            mod = importlib.import_module(module_path)
            transform_function = getattr(mod, function_name)
        except ModuleNotFoundError:
            raise InvalidFunctionError(f"Invalid module for a custom function: '{function_name}'")
        except ValueError:
            raise InvalidFunctionError(f"Invalid module for a custom function: '{function_name}'")
        except AttributeError:
            raise InvalidFunctionError(f"Invalid name for a custom function: '{function_name}'")

        return transform_function(*args_list)


from typing import BinaryIO, Iterator

from defusedxml.ElementTree import parse as parse_xml

from RecordMapper.common import Reader

class XMLReader(Reader):
    """A object that reads records form an input xml file.
    """

    def __init__(self, file_path: str):
        """The constructor of the XMLReader.

        :param input: The path of the file.
        :type input: str
        """

        super().__init__(file_path)
        self.filepath = file_path

    def read_records_from_input(self, input_stream: BinaryIO) -> Iterator[dict]:
        """The function that reads the records from the input stream.

        :param input_stream: The input stream of the records.
        :type input_stream: BinaryIO
        :yield: dict
        :rtype: 
        """

        self.reader = parse_xml(input_stream).getroot()

        for record in self.reader:
            yield {attribute.tag : attribute.text for attribute in record}
